#include <linux/arm-smccc.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/kdev_t.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>     //kmalloc()
#include <linux/uaccess.h>  //copy_to/from_user()

struct SmcData {
    uint64_t input[4];   // input data 32 byte
    uint64_t output[8];  // ouput data 64 byte
};

#define RD_RANDOM _IOWR('c', 1, struct SmcData *)

dev_t dev = 0;
static struct class *dev_class;
static struct cdev etx_cdev;

/*
** Function Prototypes
*/
static int __init etx_driver_init(void);
static void __exit etx_driver_exit(void);
static int etx_open(struct inode *inode, struct file *file);
static int etx_release(struct inode *inode, struct file *file);
static ssize_t etx_read(struct file *filp, char __user *buf, size_t len, loff_t *off);
static ssize_t etx_write(struct file *filp, const char *buf, size_t len, loff_t *off);
static long etx_ioctl(struct file *file, unsigned int cmd, unsigned long arg);

/*
** File operation sturcture
*/
static struct file_operations fops = {
    .owner = THIS_MODULE,
    .read = etx_read,
    .write = etx_write,
    .open = etx_open,
    .unlocked_ioctl = etx_ioctl,
    .release = etx_release,
};

/*
** This fuction will be called when we open the Device file
*/
static int etx_open(struct inode *inode, struct file *file)
{
    return 0;
}

/*
** This fuction will be called when we close the Device file
*/
static int etx_release(struct inode *inode, struct file *file)
{
    return 0;
}

/*
** This fuction will be called when we read the Device file
*/
static ssize_t etx_read(struct file *filp, char __user *buf, size_t len, loff_t *off)
{
    return 0;
}

/*
** This fuction will be called when we write the Device file
*/
static ssize_t etx_write(struct file *filp, const char __user *buf, size_t len, loff_t *off)
{
    return 0;
}

/*
** This fuction will be called when we write IOCTL on the Device file
*/
static long etx_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    unsigned long x0;
    struct SmcData msg;
    struct arm_smccc_res smccc_result;
    switch (cmd) {
        case RD_RANDOM:
            if (copy_from_user(&msg, (struct SmcData *) arg, sizeof(struct SmcData)) != 0) {
                printk(KERN_ERR "copyfrom error");
                return -1;
            }
            x0 = 0xC2001000;  // function_id
            // 将msg.input 0 ~ 3 作为入参输入
            arm_smccc_smc(x0, msg.input[0], msg.input[1], msg.input[2], msg.input[3], 0, 0, 0, &smccc_result);
            // 将smccc_result的结果读到msg.output 0 ~ 3 中
            msg.output[0] = smccc_result.a0;
            msg.output[1] = smccc_result.a1;
            msg.output[2] = smccc_result.a2;
            msg.output[3] = smccc_result.a3;
            x0 = 0xC2001001;  // function_id
            arm_smccc_smc(x0, 0, 0, 0, 0, 0, 0, 0, &smccc_result);
            // 将smccc_result的结果读到msg.output 4 ~ 7 中
            msg.output[4] = smccc_result.a0;
            msg.output[5] = smccc_result.a1;
            msg.output[6] = smccc_result.a2;
            msg.output[7] = smccc_result.a3;
            if (copy_to_user((struct SmcData *) arg, &msg, sizeof(struct SmcData)) != 0) {
                printk(KERN_ERR "copyto error");
                return -1;
            }
            break;
        default:
            break;
    }
    return 0;
}

int etx_driver_init(void)
{
    printk(KERN_INFO "smc driver init... ");
    /*Allocating Major number*/
    if ((alloc_chrdev_region(&dev, 0, 1, "smc_dri_DEV")) < 0) {
        return -1;
    }

    /*Creating cdev structure*/
    cdev_init(&etx_cdev, &fops);

    /*Adding character device to the system*/
    if ((cdev_add(&etx_cdev, dev, 1)) < 0) {
        goto r_class;
    }

    /*Creating struct class*/
    if ((dev_class = class_create(THIS_MODULE, "smc_dri_class")) == NULL) {
        goto r_class;
    }

    /*Creating device*/
    if ((device_create(dev_class, NULL, dev, NULL, "smc_dri")) == NULL) {
        goto r_device;
    }
    printk(KERN_INFO "done!!!\n");
    return 0;

r_device:
    class_destroy(dev_class);
r_class:
    unregister_chrdev_region(dev, 1);
    return -1;
}

void etx_driver_exit(void)
{
    printk(KERN_INFO "smc driver exit... ");
    device_destroy(dev_class, dev);
    class_destroy(dev_class);
    cdev_del(&etx_cdev);
    unregister_chrdev_region(dev, 1);
    printk(KERN_INFO "done!!!\n");
}

MODULE_LICENSE("GPL");
module_init(etx_driver_init);
module_exit(etx_driver_exit);
